import json
import logging
import os
from typing import List

import etcd3

from galileoexperiments.api.model import Pod
from galileoexperiments.utils.constants import function_label, zone_label

logger = logging.getLogger(__name__)


class EtcdClient:
    _etcd_client: etcd3

    def __init__(self, etcd_host: str, etcd_port: int):
        self.etcd_host = etcd_host
        self.etcd_port = etcd_port
        self._etcd_client = etcd3.client(host=etcd_host, port=etcd_port)

    @staticmethod
    def from_env() -> 'EtcdClient':
        etcd_host = os.environ.get('etcd_host', 'localhost')
        etcd_port = int(os.environ.get('etcd_port', 2379))
        logger.info(f"Connect to etcd instance {etcd_host}:{etcd_port}")
        return EtcdClient(etcd_host, etcd_port)

    def write(self, key: str, value: str):
        self._etcd_client.put(key, value)

    def remove(self, key: str):
        self._etcd_client.delete(key)

def set_weights_rr(pods: List[Pod], zone: str, fn: str):
    client = EtcdClient.from_env()
    weights = {
        "ips": [f'{pod.ip}:8080' for pod in pods],
        "weights": [1] * len(pods)
    }
    key = f'golb/function/{zone}/{fn}'
    value = json.dumps(weights)
    logger.info(f'Set following in etcd {key} - {value}')
    client.write(key=key, value=value)
    return key


def set_weight(pod: Pod, weight: int):
    client = EtcdClient.from_env()
    zone = pod.labels[zone_label]
    fn = pod.labels[function_label]
    ip = pod.ip
    key = f'golb/function/{zone}/{fn}'
    value = json.dumps({"ips": [f'{ip}:8080'], "weights": [weight]})
    client.write(key=key, value=value)



import pickle

import redis
from galileo.worker.api import ClientDescription


def clear_list(list_key: str, rds: redis.Redis):
    while rds.llen(list_key) != 0:
        rds.lpop(list_key)


def read_and_save_profile(profile_path: str, client_desc: ClientDescription, rds: redis.Redis):
    with open(profile_path, 'rb') as fd:
        ias = pickle.load(fd)
        for index, ia in enumerate(ias):
            if ia == 0:
                # prevents of using 0 because it may lead to crash
                ias[index] = 0.00000000001
        # TODO for some reason galileo workers skip the last entry, thus skipping the last ia
        ias.append(0.00000000001)
        list_key = client_desc.client_id
        print(f'clear list, with key: {list_key}')
        clear_list(list_key, rds)
        print('list cleared, start push')
        rds.lpush(list_key, *ias[0:])
        llen = rds.llen(list_key)
        print('pushed list')



import logging
import time

from galileo.shell.shell import RoutingTableHelper, Galileo

from galileoexperiments.api.model import ProfilingWorkloadConfiguration, \
    ExperimentRunConfiguration, AppWorkloadConfiguration, ProfilingExperimentConfiguration
from galileoexperiments.api.profiling import GalileoClientGroupConfig
from galileoexperiments.experiment.run import run_profiling_experiment
from galileoexperiments.utils.arrivalprofile import clear_list, read_and_save_profile
from galileoexperiments.utils.constants import function_label, zone_label
from galileoexperiments.utils.helpers import set_weights_rr, EtcdClient
from galileoexperiments.utils.k8s import spawn_pods, get_pods, remove_pods

logger = logging.getLogger(__name__)


def _run(config: ProfilingExperimentConfiguration):
    if config.exp_run_config.exp_name is None:
        config.exp_run_config.exp_name = f'{config.app_name}-{int(time.time())}'
    logger.info("Set up experiment")

    rtbl: RoutingTableHelper = config.rtbl
    service = f'{config.app_name}-{config.zone}'

    try:
        url = f'{config.lb_ip}:8080'
        logger.info(f"Set routing table '{service} - {url}'")
        rtbl.set(service, [url], [1])
        return run_profiling_experiment(config)
    finally:
        if rtbl is not None:
            logger.info(f"Remove routing table '{service}'")
            rtbl.remove(service)


def run_profiling_workload(workload_config: ProfilingWorkloadConfiguration):
    rds = workload_config.rds
    galileo: Galileo = workload_config.galileo
    client_group = None

    zone = workload_config.zone
    creator = workload_config.creator
    master_node = workload_config.master_node
    host = workload_config.host
    image = workload_config.image
    profiling_app = workload_config.profiling_app

    if workload_config.params.get('exp') is None or workload_config.params['exp'].get('requests') is None:
        workload_config.params['exp'] = {
            'requests': {}
        }

    use_profiles = workload_config.profiles is not None
    if use_profiles:
        profiles = workload_config.profiles
        workload_config.params['exp']['requests']['profiles'] = profiles
        n_clients = len(profiles)
        workload_config.params['exp']['requests']['n_clients'] = n_clients
        workload_config.params['exp']['requests']['no_pods'] = workload_config.no_pods

        client_group_config = GalileoClientGroupConfig(
            n_clients=n_clients,
            zone=workload_config.zone,
            fn_name=workload_config.app_name,
            params=workload_config.params
        )

        client_group = profiling_app.spawn_group(n_clients, rds, galileo, client_group_config)
        time.sleep(1)
        for index, client in enumerate(client_group.clients):
            profile_path = profiles[index]
            clear_list(client.client_id, rds)
            read_and_save_profile(profile_path, client, rds)


        def requests():
            client_group.request(ia=('prerecorded', 'ran')).wait()
            client_group.close()

        try:
            exp_run_config = ExperimentRunConfiguration(
                creator=creator,
                master_node=master_node,
                galileo_context=workload_config.context,
                metadata=workload_config.params,
            )
            app_workload_config = AppWorkloadConfiguration(
                app_container_image=image,
                pod_factory=profiling_app.pod_factory,
                requests=requests,
            )
            config = ProfilingExperimentConfiguration(
                app_name=workload_config.app_name,
                zone=zone,
                host=host,
                no_pods=workload_config.no_pods,
                n_clients=n_clients,
                app_workload_config=app_workload_config,
                exp_run_config=exp_run_config,
                lb_ip=workload_config.lb_ip
            )

            logger.info(f'run: {workload_config.params}')
            _run_profiling_experiment(config)

        except Exception as e:
            logger.error(e)

    else:
        workload_config.params['exp']['requests']['n'] = workload_config.n
        workload_config.params['exp']['requests']['ia'] = workload_config.ia
        workload_config.params['exp']['requests']['n_clients'] = workload_config.n_clients
        workload_config.params['exp']['requests']['no_pods'] = workload_config.no_pods

        client_group_config = GalileoClientGroupConfig(
            n_clients=workload_config.n_clients,
            zone=workload_config.zone,
            fn_name=workload_config.app_name,
            params=workload_config.params
        )

        client_group = profiling_app.spawn_group(workload_config.n_clients, rds, galileo, client_group_config)

        def requests():
            # FIXME for some reason galileo workers always misses one request
            client_group.request(n=workload_config.n+1, ia=workload_config.ia).wait()
            client_group.close()

        try:
            exp_run_config = ExperimentRunConfiguration(
                creator=creator,
                master_node=master_node,
                galileo_context=workload_config.context,
                metadata=workload_config.params,
            )
            app_workload_config = AppWorkloadConfiguration(
                app_container_image=image,
                pod_factory=profiling_app.pod_factory,
                requests=requests
            )
            config = ProfilingExperimentConfiguration(
                app_name=workload_config.app_name,
                zone=zone,
                host=host,
                no_pods=workload_config.no_pods,
                n_clients=workload_config.n_clients,
                app_workload_config=app_workload_config,
                exp_run_config=exp_run_config,
                lb_ip=workload_config.lb_ip
            )

            logger.info(f'run: {workload_config.params}')
            _run_profiling_experiment(config)
        except Exception as e:
            logger.error(e)


def _run_profiling_experiment(config: ProfilingExperimentConfiguration):
    pod_names = None
    params = config.exp_run_config.metadata
    image = config.app_workload_config.app_container_image
    name = config.app_name
    pod_prefix = f'{name}-deployment'
    host = config.host
    no_pods = config.no_pods
    n_clients = config.n_clients
    etcd_service_key = None
    params['exp']['host'] = host
    params['exp']['zone'] = config.zone
    params['exp']['app_name'] = config.app_name
    params['exp']['app_container_image'] = config.app_workload_config.app_container_image

    try:
        labels = {
            function_label: config.app_name,
            zone_label: config.zone
        }

        pod_names = spawn_pods(image, pod_prefix, host, labels, no_pods, config.app_workload_config.pod_factory)
        logger.info('Sleep for 5 seconds, to wait that pods are placed')
        time.sleep(5)
        pods = get_pods(pod_names)

        logger.info("Set weights for Pod(s)")
        etcd_service_key = set_weights_rr(pods, config.zone, name)

        time.sleep(1)
        if config.exp_run_config.exp_name is None:
            experiment_name = f'{name}-clients-{n_clients}-{int(time.time())}'
            config.exp_run_config.exp_name = experiment_name
        return _run(config)
    except Exception as e:
        logger.error(e)
    finally:
        if pod_names is not None:
            logger.info(f'Remove {len(pod_names)} pods')
            remove_pods(pod_names)
        if etcd_service_key is not None:
            client = EtcdClient.from_env()
            client.remove(etcd_service_key)


import logging
from dataclasses import dataclass
from typing import List, Dict, Callable

from kubernetes import client, config
from kubernetes.client import V1Deployment, V1ObjectMeta, V1DeploymentSpec, V1LabelSelector, V1PodTemplateSpec, \
    V1PodSpec, V1Toleration, V1Container, V1EnvFromSource, V1ConfigMapEnvSource
from kubernetes.client import V1ResourceRequirements, V1EnvVar, V1EnvVarSource, V1ObjectFieldSelector

from galileoexperiments.api.model import Pod

logger = logging.getLogger(__name__)


def start_telemd_kubernetes_adapter(master_node: str) -> V1Deployment:
    # Configs can be set in Configuration class directly or using helper utility
    config.load_kube_config()

    v1 = client.AppsV1Api()
    image = 'edgerun/telemd-kubernetes-adapter:0.1.18'
    return v1.create_namespaced_deployment(pretty=True, namespace='default',
                                           body=V1Deployment(
                                               api_version='apps/v1',
                                               kind='Deployment',
                                               metadata=V1ObjectMeta(name='telemd-kubernetes-adapter'),
                                               spec=V1DeploymentSpec(
                                                   replicas=1,
                                                   selector=V1LabelSelector(match_labels={
                                                       'app': 'telemd-kubernetes-adapter'
                                                   }),
                                                   template=V1PodTemplateSpec(
                                                       metadata=V1ObjectMeta(
                                                           labels={'app': 'telemd-kubernetes-adapter'}),
                                                       spec=V1PodSpec(
                                                           tolerations=[
                                                               V1Toleration(
                                                                   key='node-role.kubernetes.io/master',
                                                                   operator='Exists',
                                                                   effect='NoSchedule'
                                                               )
                                                           ],
                                                           node_selector={
                                                               'kubernetes.io/hostname': master_node
                                                           },
                                                           containers=[
                                                               V1Container(
                                                                   name='telemd-kubernetes-adapter',
                                                                   image=image,
                                                                   env_from=[
                                                                       V1EnvFromSource(
                                                                           config_map_ref=
                                                                           V1ConfigMapEnvSource(
                                                                               name='telemd-kubernetes-adapter-config'
                                                                           )
                                                                       )
                                                                   ]
                                                               )
                                                           ]
                                                       )
                                                   ),
                                               ),
                                           ))


def stop_telemd_kubernetes_adapter():
    # Configs can be set in Configuration class directly or using helper utility
    config.load_kube_config()

    v1 = client.AppsV1Api()
    v1.delete_namespaced_deployment(name='telemd-kubernetes-adapter', namespace='default')


def get_pods(pod_names: List[str]) -> List[Pod]:
    config.load_kube_config()
    v1 = client.CoreV1Api()
    pod_list = v1.list_namespaced_pod("default")
    pods = []
    for pod in pod_list.items:
        if pod.metadata.name in pod_names:
            ip = pod.status.pod_ip
            labels = pod.metadata.labels
            pod_id = pod.metadata.uid
            name = pod.metadata.name
            pods.append(Pod(pod_id, ip, labels, name))

    return pods


def spawn_pods(image: str, name: str, node: str, labels: Dict[str, str], n: int, pod_factory: Callable[[str, str, Dict], client.V1Container]) -> List[str]:
    """
    Function spawns n pods on the given node. The pod factory creates the containers to allow
    different kinds of containers.
    :param image: the container imag
    :param name: the pod name prefix
    :param node: the node
    :param labels: labels to attach
    :param n: the number of pods to spawn on the given node
    :param pod_factory: factory function to create V1Containers
    :return: a list containing the names of pods created
    """
    # Configs can be set in Configuration class directly or using helper utility
    config.load_kube_config()
    resource_requests = {}
    v1 = client.CoreV1Api()
    pods = []
    for idx in range(n):
        selector = {'kubernetes.io/hostname': node}
        pod_name = f'{name}-{node}-{idx}'
        pod = client.V1Pod(
            api_version="v1",
            kind="Pod",
            metadata=client.V1ObjectMeta(name=pod_name, labels=labels),
            spec=client.V1PodSpec(
                node_selector=selector,
                containers=[
                    pod_factory(pod_name, image, resource_requests)
                ]
            ),
        )
        logger.info(f"Create pod '{pod_name}'")
        v1.create_namespaced_pod('default', pod, async_req=False)
        pods.append(pod_name)
    return pods


def remove_pods(names: List[str]):
    config.load_kube_config()
    v1 = client.CoreV1Api()
    for name in names:
        v1.delete_namespaced_pod(name, 'default', async_req=False)



zone_label = "ether.edgerun.io/zone"
pod_type_label = "type"
function_label = "ether.edgerun.io/function"
api_gateway_type_label = 'api-gateway'
function_type_label = 'fn'
client_role_label = 'node-role.kubernetes.io/client'
worker_role_label = 'node-role.kubernetes.io/worker'

# Pod status constants
pod_not_running = 'Not Running'
pod_running = 'Running'
pod_unknown = "Unknown"
pod_failed = "Failed"
pod_succeeded = "Succeeded"
pod_pending = "Pending"



import json
from dataclasses import dataclass
from typing import Dict, Callable, Optional, List, Tuple, Union

import redis
from galileo.shell.shell import RoutingTableHelper, Galileo, Telemd, Experiment
from kubernetes import client

from galileoexperiments.api.profiling import ProfilingApplication


@dataclass
class Pod:
    pod_id: str
    ip: str
    labels: Dict[str, str]
    name: str


@dataclass
class ExperimentRunConfiguration:
    creator: str
    # the kubernetes master node
    master_node: str
    galileo_context: Dict[str, object]
    # arbitrary Dict that will be saved with the Experiment
    metadata: Optional[Dict] = None
    exp_name: str = None

    @property
    def galileo(self) -> Galileo:
        return self.galileo_context['g']

    @property
    def telemd(self) -> Telemd:
        return self.galileo_context['telemd']

    @property
    def exp(self) -> Experiment:
        return self.galileo_context['exp']

    @property
    def rtbl(self) -> RoutingTableHelper:
        return self.galileo_context['rtbl']

    @property
    def rds(self) -> redis.Redis:
        return self.galileo_context['rds']



@dataclass
class ProfilingWorkloadConfiguration:
    """
    This class encapsulates the workload of one profiling experiment.
    Either you pass a list of profiles, or configure the remaining parameters.
    """
    creator: str
    # the application name
    app_name: str
    # the host that will be profiled
    host: str
    # load balancer ip that forwards requests
    lb_ip: str
    # the zone in which the host is
    zone: str
    # the kubernetes master hostname
    master_node: str
    # the container image
    image: str
    # number of pods
    no_pods: int
    # stores params for service and metadata about experiment
    params: Dict
    # specifies how to create a ClientGroup and V1Containers
    profiling_app: ProfilingApplication
    # context contains all dependencies, instantiated using `galileo.shell.shell.init`
    context: Dict
    # list of profiles - one per client
    profiles: List[str] = None
    # number of requests
    n: int = None
    # interarrival config
    ia: Union[float, Tuple] = None
    # number of clients
    n_clients: int = None

    @property
    def galileo(self) -> Galileo:
        return self.context['g']

    @property
    def telemd(self) -> Telemd:
        return self.context['telemd']

    @property
    def exp(self) -> Experiment:
        return self.context['exp']

    @property
    def rtbl(self) -> RoutingTableHelper:
        return self.context['rtbl']

    @property
    def rds(self) -> redis.Redis:
        return self.context['rds']


@dataclass
class ScenarioWorkloadConfiguration:
    """
    This class encapsulates the workload of one profiling experiment.
    Either you pass a list of profiles, or configure the remaining parameters.
    """
    creator: str

    # image: the application names
    app_names: Dict[str, str]

    # the kubernetes master hostname
    master_node: str

    # {node: { image: number of instances}
    services: Dict[str, Dict[str, int]]

    # {node: zone}
    zone_mapping: Dict[str, str]

    # stores general experiment parameters
    params: Dict

    # stores specific app params (by image)
    app_params: Dict[str, Dict]

    # per image an app that specifies how to create a ClientGroup and V1Containers
    profiling_apps: Dict[str, ProfilingApplication]

    # context contains all dependencies, instantiated using `galileo.shell.shell.init`
    context: Dict

    # load balancer ip addresses per zone
    lb_ips: Dict[str, str]

    # per zone: {image: list of profiles - one per client}
    profiles: Dict[str, Dict[str, List[str]]]

    @property
    def galileo(self) -> Galileo:
        return self.context['g']

    @property
    def telemd(self) -> Telemd:
        return self.context['telemd']

    @property
    def exp(self) -> Experiment:
        return self.context['exp']

    @property
    def rtbl(self) -> RoutingTableHelper:
        return self.context['rtbl']

    @property
    def rds(self) -> redis.Redis:
        return self.context['rds']


@dataclass
class AppWorkloadConfiguration:
    app_container_image: str
    requests: Callable[[], None]
    pod_factory: Callable[[str, str], client.V1Container]


@dataclass
class ProfilingExperimentConfiguration:
    app_name: str
    # used to determine the zone in which the clients will spawn
    zone: str
    # hosts that should emit telemetry data and is being profiled
    host: str
    # load balancer ip that forwards requests
    lb_ip: str
    # how many pods should be started on the host
    no_pods: int
    # how many clients should be started
    n_clients: int
    app_workload_config: AppWorkloadConfiguration
    exp_run_config: ExperimentRunConfiguration

    @property
    def rtbl(self) -> RoutingTableHelper:
        return self.exp_run_config.rtbl

@dataclass
class ScenarioExperimentConfiguration:
    # apps per (function, zone)
    apps: List[AppWorkloadConfiguration]
    exp_run_config: ExperimentRunConfiguration




import logging
import time
from typing import Callable

from galileoexperiments.api.model import ProfilingExperimentConfiguration, ScenarioExperimentConfiguration, \
    ExperimentRunConfiguration
from galileoexperiments.utils.k8s import start_telemd_kubernetes_adapter, stop_telemd_kubernetes_adapter
from galileoexperiments.utils.rds import wait_for_galileo_events

logger = logging.getLogger(__name__)


def run_profiling_experiment(config: ProfilingExperimentConfiguration):
    run_experiment(config.exp_run_config, config.app_workload_config.requests)


def run_scenario_experiment(config: ScenarioExperimentConfiguration, requests: Callable):
    return run_experiment(config.exp_run_config, requests)


def run_experiment(config: ExperimentRunConfiguration, requests: Callable):
    metadata = config.metadata
    if metadata is None:
        metadata = {}
    try:

        # discover workers
        workers = config.galileo.discover()
        logger.info(f"Discovered workers: {workers}")
        time.sleep(1)

        # toggle tracing
        logger.info("Start tracing")
        config.galileo.start_tracing()

        # unpause telemd
        logger.info(f"Unpause telemd")
        config.telemd.start_telemd()

        # start exp
        logger.info("Start experiment and wait for 1 second")
        config.exp.start(
            name=config.exp_name,
            creator=config.creator,
            metadata=metadata
        )
        time.sleep(1)

        # start telemd kubernetes adapter
        start_telemd_kubernetes_adapter(config.master_node)
        logger.info("Waiting for telemd_kubernetes adapter to publish galileo events")
        wait_for_galileo_events(config.rds)
        time.sleep(1)

        # set requests
        logger.info("start requests")
        requests()
        time.sleep(5)

    except Exception as e:
        logger.error(e)
    finally:
        logger.info("Stop tracing")
        config.galileo.stop_tracing()
        logger.info("Pause telemd")
        config.telemd.stop_telemd()
        logger.info("Stop exp")
        config.exp.stop()
        logger.info("Shutdown telemd kubernetes adapter")
        stop_telemd_kubernetes_adapter()



import abc
from dataclasses import dataclass
from typing import Dict

import redis
from galileo.shell.shell import Galileo, ClientGroup
from kubernetes import client

@dataclass
class GalileoClientGroupConfig:
    n_clients: int
    zone: str
    fn_name: str
    # misc. paremeters for service
    params: Dict

class ProfilingApplication(abc.ABC):

    def spawn_group(self, clients: int, rds: redis.Redis, galileo: Galileo,
                    config: GalileoClientGroupConfig) -> ClientGroup: ...

    def pod_factory(self, pod_name: str, image: str, resource_requests: Dict) -> client.V1Container: ...



import logging
import time
from typing import List, Dict, Tuple, Callable

from galileo.shell.shell import RoutingTableHelper, ClientGroup

from galileoexperiments.api.model import ScenarioWorkloadConfiguration, Pod, ScenarioExperimentConfiguration, \
    ExperimentRunConfiguration, AppWorkloadConfiguration
from galileoexperiments.api.profiling import GalileoClientGroupConfig
from galileoexperiments.experiment.run import run_scenario_experiment
from galileoexperiments.utils.arrivalprofile import clear_list, read_and_save_profile
from galileoexperiments.utils.constants import function_label, zone_label
from galileoexperiments.utils.helpers import set_weights_rr, EtcdClient
from galileoexperiments.utils.k8s import spawn_pods, get_pods, remove_pods

logger = logging.getLogger(__name__)


def spawn_pods_for_config(workload_config: ScenarioWorkloadConfiguration) -> List[Pod]:
    pod_names = []
    for host, values in workload_config.services.items():
        for image, no_pods in values.items():
            name = workload_config.app_names[image]

            labels = {
                function_label: name,
                zone_label: workload_config.zone_mapping[host]
            }
            profiling_app = workload_config.profiling_apps[image]
            pod_name_prefix = f'{name}-deployment'
            names = spawn_pods(image, pod_name_prefix, host, labels, no_pods, profiling_app.pod_factory)
            pod_names.extend(names)
    # TODO remove sleep and implement approach to continuously poll for pods till their IP is available
    time.sleep(5)
    return get_pods(pod_names)


def _map_pods_to_dict(pods: List[Pod]) -> Dict[Tuple[str, str], List[Pod]]:
    pods_map = {}
    for pod in pods:
        zone = pod.labels[zone_label]
        fn = pod.labels[function_label]
        if pods_map.get((fn, zone)) is None:
            pods_map[(fn, zone)] = [pod]
        else:
            pods_map[(fn, zone)].append(pod)

    return pods_map


def set_loadbalancer_weights(pods: Dict[Tuple[str, str], List[Pod]]):
    keys = []
    for t, pods in pods.items():
        fn = t[0]
        zone = t[1]
        key = set_weights_rr(pods, zone, fn)
        keys.append(key)
    return keys


def set_rtbl(fns: List[str], load_balancers: Dict[str, str], rtbl: RoutingTableHelper) -> List[str]:
    """
    :param fns: function names available
    :param load_balancers: load balancer ip addresses by zone
    """
    services = []
    for fn in fns:
        for zone, lb_ip in load_balancers.items():
            service = f'{fn}-{zone}'
            url = f'{lb_ip}:8080'
            logger.info(f"Set routing table '{service} - {url}'")
            rtbl.set(service, [url], [1])
            services.append(service)
    return services


def prepare_client_groups_for_services(workload_config: ScenarioWorkloadConfiguration) -> Tuple[List[
                                                                                                    Tuple[
                                                                                                        str, str, ClientGroup]], Callable]:
    client_groups = []
    for zone, values in workload_config.profiles.items():
        for image, profiles in values.items():

            n_clients = len(profiles)
            client_group_config = GalileoClientGroupConfig(
                n_clients=n_clients,
                zone=zone,
                fn_name=workload_config.app_names[image],
                params=workload_config.app_params[image]
            )
            profiling_app = workload_config.profiling_apps[image]
            rds = workload_config.rds
            galileo = workload_config.galileo
            client_group = profiling_app.spawn_group(n_clients, rds, galileo, client_group_config)
            time.sleep(1)
            for index, client in enumerate(client_group.clients):
                profile_path = profiles[index]
                clear_list(client.client_id, rds)
                read_and_save_profile(profile_path, client, rds)

            client_groups.append((image, zone, client_group))

    def requests():
        for idx, group in enumerate(client_groups):
            if idx == len(client_groups) - 1:
                # TODO the wait here might return sooner than other profiles => implement solution that waits for all
                group[2].request(ia=('prerecorded', 'ran')).wait()
            else:
                group[2].request(ia=('prerecorded', 'ran'))

    return client_groups, requests


def set_params(workload_config: ScenarioWorkloadConfiguration):
    workload_config.params['app_params'] = workload_config.app_params
    workload_config.params['profiles'] = workload_config.profiles
    workload_config.params['lb_ips'] = workload_config.lb_ips
    workload_config.params['zone_mapping'] = workload_config.zone_mapping
    workload_config.params['services'] = workload_config.services
    workload_config.params['app_names'] = workload_config.app_names


def run_scenario_workload(workload_config: ScenarioWorkloadConfiguration):
    rtbl: RoutingTableHelper = workload_config.rtbl
    pods = None
    rtbl_services = []
    etcd_service_keys = []
    creator = workload_config.creator
    master_node = workload_config.master_node
    client_groups = []
    try:
        client_groups, requests = prepare_client_groups_for_services(workload_config)
        pods = spawn_pods_for_config(workload_config)
        etcd_service_keys = set_loadbalancer_weights(_map_pods_to_dict(pods))
        rtbl_services = set_rtbl(list(workload_config.app_names.values()), workload_config.lb_ips, rtbl)


        set_params(workload_config)

        exp_run_config = ExperimentRunConfiguration(
            creator=creator,
            master_node=master_node,
            galileo_context=workload_config.context,
            metadata=workload_config.params,
        )
        app_configs = []

        for (image, zone, client_group) in client_groups:
            app_workload_config = AppWorkloadConfiguration(
                app_container_image=image,
                requests=lambda x: None,
                pod_factory=workload_config.profiling_apps[image].pod_factory
            )
            app_configs.append(app_workload_config)

        scenario_experiment_config = ScenarioExperimentConfiguration(
            apps=app_configs,
            exp_run_config=exp_run_config
        )
        run_scenario_experiment(scenario_experiment_config, requests)
    except Exception as e:
        logger.error(e)
    finally:
        if pods is not None:
            logger.info(f'Remove {len(pods)} pods')
            remove_pods([x.name for x in pods])
        for service in rtbl_services:
            logger.info(f'Remove rtbl entry for: {service}')
            rtbl.remove(service)
        client = EtcdClient.from_env()
        for key in etcd_service_keys:
            client.remove(key)
        for c_group in client_groups:
            c_group[2].close()



import os

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements-dev.txt", "r") as fh:
    tests_require = [line for line in fh.read().split(os.linesep) if line]

with open("requirements.txt", "r") as fh:
    install_requires = [line for line in fh.read().split(os.linesep) if line]

setuptools.setup(
    name="edgerun-galileo-experiments",
    version="0.0.2.dev4",
    author="Philipp Raith, Thomas Rausch",
    author_email="p.raith@dsg.tuwien.ac.at, t.rausch@dsg.tuwien.ac.at",
    description="Galileo Experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/edgerun/galileo-experiments",
    packages=setuptools.find_packages(),
    setup_requires=['wheel'],
    install_requires=install_requires,
    pyton_requires='>=3.9',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={
    },

)
